
(* configuration *)
let mindim = 1
let maxdim = 1
let trynum = 20

(* ---------------------------------------------------------------------- *)

(* decode terminary streams  *)

let rec pow x y =
  if y == 0 then 1
  else x * pow x (pred y)
;;

let maxmatrixcode = ref 0
let get_matrixcode n = 
  let matrixcode = [| 
    "state.A[0]";
    "state.A[1]";
    "state.A[2]";
    "state.A[3]";
    "state.A[4]";
    "state.A[5]";
  |] in
  if !maxmatrixcode < n then
    maxmatrixcode := n;
  matrixcode.(n)
;;

let rec parse_int str rpos = 
  match str.[!rpos] with
    | '0' ->
        incr rpos; 0
    | '1' -> 
        incr rpos;
        let answer = 1 + (Char.code str.[!rpos] - Char.code '0') in
        incr rpos;
        answer
    | '2' ->
        incr rpos; assert (str.[!rpos] = '2'); incr rpos;
        let x = parse_int str rpos + 2 in
        let alpha = 
          let rec loop sum i = 
            if i < x then
              loop (sum + pow 3 i) (succ i)
            else 
              sum
          in loop 0 0
        in
        let tt = String.sub str !rpos x in
        let beta = 
          let rec loop sum i =
            if i < String.length tt then
              loop (sum * 3 + (Char.code tt.[i] - Char.code '0')) (succ i)
            else
              sum
          in loop 0 0
        in
        rpos := !rpos + x;
        let answer = alpha + beta in
        answer
    | _ -> failwith "unknown token."
;;

(* unit test *)
let _ = begin
  assert (parse_int "0" (ref 0) = 0);
  assert (parse_int "10" (ref 0) = 1);
  assert (parse_int "11" (ref 0) = 2);
  assert (parse_int "12" (ref 0) = 3);
  assert (parse_int "22000" (ref 0) = 4);
  assert (parse_int "22001" (ref 0) = 5);
  assert (parse_int "22002" (ref 0) = 6);
  assert (parse_int "22010" (ref 0) = 7);
  assert (parse_int "22011" (ref 0) = 8);
  assert (parse_int "22012" (ref 0) = 9);
  assert (parse_int "22020" (ref 0) = 10);
  assert (parse_int "22021" (ref 0) = 11);
  assert (parse_int "22022" (ref 0) = 12);
  assert (parse_int "2210000" (ref 0) = 13);
  assert (parse_int "2210222" (ref 0) = 39);
  assert (parse_int "22110000" (ref 0) = 40);
  assert (parse_int "22110001" (ref 0) = 41);
  assert (parse_int "22112222" (ref 0) = 120);
  assert (parse_int "221200000" (ref 0) = 121);
  assert (parse_int "221222222" (ref 0) = 363);
  assert (parse_int "2222000000000" (ref 0) = 364);
end

let parse_list_num str rpos: int =
  match str.[!rpos] with
    | '0' -> incr rpos; 0
    | '1' -> incr rpos; 1
    | '2' ->
        incr rpos; assert (str.[!rpos] == '2'); incr rpos;
        2 + parse_int str rpos
    | _ -> failwith "unknown"
;;

(* unit test *)
let _ = begin
  assert (parse_list_num "0" (ref 0)    = 0);
  assert (parse_list_num "000" (ref 0)    = 0);
  assert (parse_list_num "011" (ref 0)    = 0);
  assert (parse_list_num "1" (ref 0)    = 1);
  assert (parse_list_num "220" (ref 0)  = 2);
  assert (parse_list_num "2210" (ref 0) = 3);
  assert (parse_list_num "2211" (ref 0) = 4);
  assert (parse_list_num "2212" (ref 0) = 5);
  assert (parse_list_num "2222000" (ref 0) = 6);
end

let parse_list str rpos: string =
  let ans = ref [] in
  let n: int = parse_list_num str rpos in
  let rec loop i =
    if i < n then begin
      let t = parse_int str rpos in
      ans := !ans @ [get_matrixcode t];
      loop (succ i)
    end
  in loop 0;
  String.concat " * " !ans
;;

let parse_main_or_aux str rpos: bool = 
  match parse_int str rpos with
    | 0 -> true
    | 1 -> false
    | _ -> failwith "parsing error: main or aux"
;;

let decode str out_chan = 
  let rpos = ref 0 in
  let n = parse_list_num str rpos in
  Printf.fprintf out_chan "/* Chamber num: %d */\n" n;
  let rec loop i = 
    if i < n then begin
      Printf.fprintf out_chan "/* chamber #%d */\n" i;
      let upper = parse_list str rpos in
      let main  = parse_main_or_aux str rpos in
      let lower = parse_list str rpos in
      if main then begin
        Printf.fprintf out_chan "ans += Q(%s, %s);\n" upper lower
      end else begin
        Printf.fprintf out_chan "ans += P(%s, %s);\n" upper lower
      end;

      loop (succ i)
    end
  in loop 0;
;;

let solve () = 
  prerr_endline "unittest OK";

  let str = read_line () in
  
  (* *)
  let out_chan = open_out "calc_tmp2.cc" in
  decode str out_chan;
  close_out out_chan;

  let size = !maxmatrixcode in

  for dim = mindim to maxdim do
    Printf.fprintf stderr "*** SOLVING DIMENSION %d\n" dim;
    let out_chan = open_out "calc_tmp1.cc" in
    Printf.fprintf out_chan "const int MNUM = %d;\n" (size + 1);
    Printf.fprintf out_chan "const int MSIZE = %d;\n" dim;
    close_out out_chan;

    (* call g++ *) 
    let compiled = Sys.command("g++ -O3 -Wall -Wextra annealing.cc -o annealing") in
    if compiled != 0 then
      failwith "compilation failure";
    
    for i = 1 to trynum do
      (* run *)
      let solved = (try Sys.command("annealing") with _ -> 1) in
      if solved = 0 then begin
        prerr_endline "solving OK!";
        raise Exit
      end else begin
        prerr_endline "solving failure.";
      end
    done;

  done;
;;

let _ =
  try
    solve();
    prerr_endline "COULDNOT SOLVED.";
    exit (-1)
  with
    | Exit -> (* seems OK *) ()
;;

#use "tencode.ml"
#use "fuel.ml"

let _ = print_fuel fuel



